@@ -1,5 +1,16 @@
-/* Copyright (C) 2006 G.P. Halkes
-   Licensed under the Open Software License version 2.0 */
+/* Copyright (C) 2006-2007 G.P. Halkes
+   This program is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License version 3, as
+   published by the Free Software Foundation.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+*/
 
 #include <stdlib.h>
 #include <string.h>
@@ -8,20 +19,24 @@
 #include "definitions.h"
 
 static const char resetColor[] = "\033[0m";
+static unsigned int oldLineNumber = 1, newLineNumber = 1;
+static bool lastWasLinefeed = true;
 
 /** Start the diff program. */
 static FILE *startDiff(void) {
 	FILE *diff;
 	char *command;
 	size_t length;
-	
+
 	/* Allocate memory to hold the diff command with arguments. */
 	length = strlen(option.ignoreCase ? DIFF_COMMAND_CI : DIFF_COMMAND) + 1 + strlen(option.oldFile.tokens->name) + 1 + strlen(option.newFile.tokens->name) + 1;
 	if (option.diffOption != NULL)
 		length += strlen(option.diffOption) + 1;
-	if ((command = malloc(length)) == NULL) {
-		fatal(_("Could not allocate memory: %s\n"), strerror(errno));
-	}
+
+	errno = 0;
+	if ((command = malloc(length)) == NULL)
+		fatal(_("Could not allocate memory: %s\n"), errno ? strerror(errno) : _("Out of memory"));
+
 	/* Compose the diff command. */
 	strcpy(command, option.ignoreCase ? DIFF_COMMAND_CI : DIFF_COMMAND);
 	strcat(command, " ");
@@ -32,7 +47,7 @@ static FILE *startDiff(void) {
 	strcat(command, option.oldFile.tokens->name);
 	strcat(command, " ");
 	strcat(command, option.newFile.tokens->name);
-	
+
 	/* Start diff */
 	if ((diff = popen(command, "r")) == NULL)
 		fatal(_("Failed to execute diff: %s\n"), strerror(errno));
@@ -40,7 +55,31 @@ static FILE *startDiff(void) {
 	return diff;
 }
 
-typedef enum {ADD, DEL, COMMON} Mode;
+/* Note: ADD should be 0, OLD_COMMON should be DEL + COMMON. */
+typedef enum {ADD, DEL, COMMON, OLD_COMMON} Mode;
+
+/** If the last character printed was a newline, do some special handling.
+	@param mode What kind of output is generated next.
+*/
+static void doPostLinefeed(Mode mode) {
+	if (lastWasLinefeed) {
+		if (mode & COMMON)
+			mode = COMMON;
+
+		lastWasLinefeed = false;
+		if (option.lineNumbers) {
+			if (option.colorMode && mode != COMMON)
+				writeString(resetColor, sizeof(resetColor) - 1);
+			printLineNumbers(oldLineNumber, newLineNumber);
+		}
+		if (option.colorMode && option.needStartStop && mode != COMMON) {
+			if (mode == ADD)
+				writeString(option.addColor, option.addColorLen);
+			else
+				writeString(option.delColor, option.delColorLen);
+		}
+	}
+}
 
 /** Skip or print the next bit of whitespace from @a file.
 	@param file The file with whitespace.
@@ -53,14 +92,86 @@ static void handleNextWhitespace(TempFile *file, bool print, Mode mode) {
 		if (c == option.whitespaceDelimiter)
 			return;
 		if (print) {
+			doPostLinefeed(mode);
+
 			/* Less mode also over-strikes whitespace */
-			if (option.less && c != '\n') {
-				if (mode == DEL) {
-					putchar('_');
-					putchar('\010');
+			if (option.less && c != '\n' && mode == DEL) {
+				addchar('_', mode & COMMON);
+				addchar('\010', mode & COMMON);
+			}
+			addchar(c, mode & COMMON);
+
+			if (c == '\n')
+				lastWasLinefeed = true;
+		}
+
+		if (c == '\n') {
+			switch (mode) {
+				case COMMON:
+				case ADD:
+					newLineNumber++;
+					break;
+				case OLD_COMMON:
+				case DEL:
+					oldLineNumber++;
+					break;
+				default:
+					PANIC();
+			}
+		}
+	}
+}
+
+/** Skip or print the next bit of whitespace from the new file, keeping the
+	old file synchronized as far as line numbers are concerned.
+
+	This is only for printing common text, and will skip over the old
+	whitespace.
+*/
+static void handleSynchronizedNextWhitespace(void) {
+	bool oldValid = true;
+	int c;
+
+	while ((c = getc(option.newFile.whitespace->file)) != EOF) {
+		if (c == option.whitespaceDelimiter)
+			break;
+		if (option.printCommon) {
+			doPostLinefeed(COMMON);
+
+			addchar(c, true);
+
+			if (c == '\n')
+				lastWasLinefeed = true;
+		}
+
+		/* If a newline was found, see if the old file also has a newline. */
+		if (c == '\n') {
+			newLineNumber++;
+			if (oldValid) {
+				oldValid = false;
+				while ((c = getc(option.oldFile.whitespace->file)) != EOF) {
+					if (c == option.whitespaceDelimiter) {
+						oldValid = false;
+						break;
+					}
+
+					oldValid = true;
+					if (c == '\n') {
+						oldLineNumber++;
+						break;
+					}
 				}
 			}
-			putchar(c);
+		}
+	}
+
+	/* Process any remaining whitespace from the old file. */
+	if (oldValid) {
+		while ((c = getc(option.oldFile.whitespace->file)) != EOF) {
+			if (c == option.whitespaceDelimiter)
+				break;
+			if (c == '\n')
+				oldLineNumber++;
 		}
 	}
 }
@@ -75,44 +186,73 @@ static void handleNextToken(TempFile *file, bool print, Mode mode) {
 	while ((c = getc(file->file)) != EOF) {
 		if (c == '\n')
 			return;
-		if (print) {
-			/* Re-transliterate the characters, if necessary. */
-			if (option.transliterate && c == '\\') {
-				c = getc(file->file);
-				switch (c) {
-					case 'n':
-						c = '\n';
-						break;
-					case '\\':
-						break;
+
+		/* Re-transliterate the characters, if necessary. */
+		if (option.transliterate && c == '\\') {
+			c = getc(file->file);
+			switch (c) {
+				case 'n':
+					c = '\n';
+					break;
+				case '\\':
+					break;
 #ifdef NO_MINUS_A
-					default:
-						if (isxdigit(c)) {
-							int value = (c <= '9' ? c - '0' : c - 'A' + 10) << 4;
-							c = getc(file->file);
-							if (c == EOF)
-								fatal(_("Error reading back input\n"));
-							c = value + (c <= '9' ? c - '0' : c - 'A');
-						} else {
+				default:
+					if (isxdigit(c)) {
+						int value = (c <= '9' ? c - '0' : c - 'A' + 10) << 4;
+						c = getc(file->file);
+						if (c == EOF)
 							fatal(_("Error reading back input\n"));
-						}
-#else
-					default:
+						c = value + (c <= '9' ? c - '0' : c - 'A');
+					} else {
 						fatal(_("Error reading back input\n"));
+					}
+#else
+				default:
+					fatal(_("Error reading back input\n"));
 #endif
-				}
 			}
+		}
+
+		if (print) {
+			doPostLinefeed(mode);
+
 			/* Printer mode and less mode do special stuff, per character. */
 			if ((option.printer || option.less) && c != '\n') {
 				if (mode == DEL) {
-					putchar('_');
-					putchar('\010');
+					addchar('_', mode & COMMON);
+					addchar('\010', mode & COMMON);
 				} else if (mode == ADD) {
-					putchar(c);
-					putchar('\010');
+					addchar(c, mode & COMMON);
+					addchar('\010', mode & COMMON);
 				}
 			}
-			putchar(c);
+			addchar(c, mode & COMMON);
+
+			if (c == '\n')
+				lastWasLinefeed = true;
+		}
+
+		if (c == '\n') {
+			switch (mode) {
+				/* When the newline is a word character rather than a whitespace
+				   character, we can safely count old and new lines together
+				   for common words. This will keep the line numbers in synch
+				   for these cases. Note that this also means that for
+				   OLD_COMMON the line counter is not incremented. */
+				case COMMON:
+					oldLineNumber++;
+				case ADD:
+					newLineNumber++;
+					break;
+				case DEL:
+					oldLineNumber++;
+					break;
+				case OLD_COMMON:
+					break;
+				default:
+					PANIC();
+			}
 		}
 	}
 }
@@ -135,7 +275,13 @@ static void handleWord(InputFile *file, int idx, bool print, Mode mode) {
 	@param idx The last word to print (or skip).
 */
 void printToCommonWord(int idx) {
-	handleWord(&option.newFile, idx, option.printCommon, COMMON);
+	while (option.newFile.lastPrinted < idx) {
+		handleSynchronizedNextWhitespace();
+		handleNextToken(option.newFile.tokens, option.printCommon, COMMON);
+		handleNextToken(option.oldFile.tokens, false, OLD_COMMON);
+		option.newFile.lastPrinted++;
+		option.oldFile.lastPrinted++;
+	}
 }
 
 /** Print (or skip if the user doesn't want to see) words from @a file.
@@ -144,40 +290,48 @@ void printToCommonWord(int idx) {
 	@param mode Either ADD or DEL, uded for printing of start/stop markers.
 */
 static void printWords(int *range, InputFile *file, bool print, Mode mode) {
-	/* Skip to bit to print */
-	handleWord(file, range[0] - 1, false, mode);
-	
-	/* Print the firs word. As we need to add the markers AFTER the first bit of
+	ASSERT(file->lastPrinted == range[0] - 1);
+
+	/* Print the first word. As we need to add the markers AFTER the first bit of
 	   white space, we can't just ust handleWord */
 
 	/* Print preceding whitespace. Should not be overstriken, so print as common */
-	handleNextWhitespace(file->whitespace, print, COMMON);
+	handleNextWhitespace(file->whitespace, print, mode + COMMON);
+	/* Ensure that the the line numbers etc. get printed before the markers */
+	if (print)
+		doPostLinefeed(COMMON);
+
 	/* Print start marker */
 	if (print && option.needStartStop) {
 		if (mode == ADD) {
 			if (option.colorMode)
-				fwrite(option.addColor, 1, option.addColorLen, stdout);
-			fwrite(option.addStart, 1, option.addStartLen, stdout);
+				writeString(option.addColor, option.addColorLen);
+			writeString(option.addStart, option.addStartLen);
 		} else {
 			if (option.colorMode)
-				fwrite(option.delColor, 1, option.delColorLen, stdout);
-			fwrite(option.delStart, 1, option.delStartLen, stdout);
+				writeString(option.delColor, option.delColorLen);
+			writeString(option.delStart, option.delStartLen);
 		}
 	}
+
 	/* Print first word */
 	handleNextToken(file->tokens, print, mode);
 	file->lastPrinted++;
 	/* Print following words */
 	handleWord(file, range[range[1] >= 0 ? 1: 0], print, mode);
+
+	if (print)
+		doPostLinefeed(mode);
+
 	/* Print stop marker */
 	if (print && option.needStartStop) {
 		if (mode == ADD)
-			fwrite(option.addStop, 1, option.addStopLen, stdout);
+			writeString(option.addStop, option.addStopLen);
 		else
-			fwrite(option.delStop, 1, option.delStopLen, stdout);
+			writeString(option.delStop, option.delStopLen);
 
 		if (option.colorMode)
-			fwrite(resetColor, 1, sizeof(resetColor) - 1, stdout);
+			writeString(resetColor, sizeof(resetColor) - 1);
 	}
 }
 
@@ -200,8 +354,9 @@ void printEnd(void) {
 	if (!option.printCommon)
 		return;
 	while(!feof(option.newFile.tokens->file)) {
-		handleNextWhitespace(option.newFile.whitespace, true, COMMON);
+		handleSynchronizedNextWhitespace();
 		handleNextToken(option.newFile.tokens, true, COMMON);
+		handleNextToken(option.oldFile.tokens, false, OLD_COMMON);
 	}
 }
 
@@ -209,12 +364,12 @@ void printEnd(void) {
 void doDiff(void) {
 	int c, command, currentIndex, range[4];
 	FILE *diff;
-	
+
 	diff = startDiff();
 
 	if (option.needMarkers)
 		puts("======================================================================");
-	
+
 	while ((c = getc(diff)) != EOF) {
 		if (c == '<' || c == '>' || c == '-') {
 			/* Skip all lines showing differences */
@@ -261,8 +416,10 @@ void doDiff(void) {
 				printDeletedWords(range);
 			if (command != 'd')
 				printAddedWords(range + 2);
-			if (option.needMarkers)
+			if (option.needMarkers) {
 				puts("\n======================================================================");
+				lastWasLinefeed = true;
+			}
 
 			/* Update statistics */
 			switch (command) {
