@@ -1,12 +1,23 @@
-/* Copyright (C) 2006 G.P. Halkes
-   Licensed under the Open Software License version 2.0 */
+/* Copyright (C) 2006-2007 G.P. Halkes
+   This program is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License version 3, as
+   published by the Free Software Foundation.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+*/
 
 #include <ctype.h>
 
 #include "definitions.h"
 #include "optionMacros.h"
 
-#define VERSION_STRING "1.2.1"
+#define VERSION_STRING "1.3"
 
 typedef struct {
 	const char *name;
@@ -176,9 +187,13 @@ N_("-y <string>, --start-insert=<string>   String to mark begin of inserted text
 N_("-z <string>, --stop-insert=<string>    String to mark end of inserted text\n"),
 N_("-D <option>, --diff-option=<option>    Add <option> to the diff command-line\n"),
 N_("-c[<spec>], --color[=<spec>]           Color mode\n"),
+N_("-L[<width>], --line-numbers[<width>]   Prepend line numbers\n"),
+N_("-C<num>, --context=<num>               Show <num> lines of context\n"),
 NULL};
 
 PARSE_FUNCTION(parseCmdLine)
+	bool whitespaceSet = false;
+
 	int noOptionCount = 0;
 	/* Initialise options to correct values */
 	memset(&option, 0, sizeof(option));
@@ -197,15 +212,24 @@ PARSE_FUNCTION(parseCmdLine)
 	OPTIONS
 		OPTION('d', "delimiters", REQUIRED_ARG)
 			size_t i, length = parseEscapes(optarg, "delimiters");
-			for (i = 0; i < length; i++)
+			for (i = 0; i < length; i++) {
 				SET_BIT(option.delimiters, optarg[i]);
+				if (whitespaceSet && TEST_BIT(option.whitespace, optarg[i]))
+					fatal(_("Whitespace and delimiter sets overlap\n"));
+				else
+					RESET_BIT(option.whitespace, optarg[i]);
+			}
 		END_OPTION
 		OPTION('W', "white-space", REQUIRED_ARG)
+			whitespaceSet = true;
 			size_t i, length = parseEscapes(optarg, "whitespace");
-			memset(option.whitespace, 0, 64);
-			for (i = 0; i < length; i++)
+			memset(option.whitespace, 0, BITMASK_SIZE);
+			for (i = 0; i < length; i++) {
 				SET_BIT(option.whitespace, optarg[i]);
-			for (i = 0; i < 64; i++) {
+				if (TEST_BIT(option.delimiters, optarg[i]))
+					fatal(_("Whitespace and delimiter sets overlap\n"));
+			}
+			for (i = 0; i < BITMASK_SIZE; i++) {
 				if (option.whitespace[i] != (char) 0xff) {
 					int j;
 					for (j = 0; j < 7; j++)
@@ -215,7 +239,7 @@ PARSE_FUNCTION(parseCmdLine)
 					break;
 				}
 			}
-			if (i == 64)
+			if (i == BITMASK_SIZE)
 				fatal(_("All characters were designated as whitespace\n"));
 		END_OPTION
 
@@ -233,7 +257,7 @@ PARSE_FUNCTION(parseCmdLine)
 				   - If the (C) symbol (that is the c in a circle) is not available,
 					 leave as it as is. (Unicode code point 0x00A9)
 				   - G.P. Halkes is name and should be left as is. */
-				_("Copyright (C) 2006 G.P. Halkes\nLicensed under the Open Software License version 2.0\n"), stdout);
+				_("Copyright (C) 2006-2007 G.P. Halkes\nLicensed under the GNU General Public License version 3\n"), stdout);
 			exit(EXIT_SUCCESS);
 		END_OPTION
 		OPTION('1', "no-deleted", NO_ARG)
@@ -389,6 +413,17 @@ PARSE_FUNCTION(parseCmdLine)
 			option.addColorLen = strlen(option.addColor);
 			noDefaultMarkers();
 		END_OPTION
+		OPTION('L', "line-numbers", OPTIONAL_ARG)
+			if (optarg != NULL)
+				PARSE_INT(option.lineNumbers, 1, INT_MAX);
+			else
+				option.lineNumbers = DEFAULT_LINENUMBER_WIDTH;
+		END_OPTION
+		OPTION('C', "context", REQUIRED_ARG)
+			PARSE_INT(option.contextLines, 0, INT_MAX);
+			option.context = true;
+			initContextBuffers();
+		END_OPTION
 		fatal(_("Option %.*s does not exist\n"), (int) optlength, CURRENT_OPTION);
 	NO_OPTION
 		switch (noOptionCount++) {
@@ -405,9 +440,6 @@ PARSE_FUNCTION(parseCmdLine)
 	/* Check that we have something to work with. */
 	if (noOptionCount != 2)
 		fatal(_("Need two files to compare\n"));
-	/* Check that the delimiters and the whitespace are disjunct. */
-	if (strpbrk(option.whitespace, option.delimiters) != NULL)
-		fatal(_("Whitespace and delimiter sets overlap\n"));
 	
 	/* Check and set some values */
 	if (option.delStart == NULL) {
@@ -427,16 +459,16 @@ PARSE_FUNCTION(parseCmdLine)
 		option.addStopLen = 2;
 	}
 
-	if (!option.printCommon && !option.printAdded && !option.printDeleted) {
+	if (!option.printCommon && !option.printAdded && !option.printDeleted)
 		option.needMarkers = false;
-	}
+
 	if ((!option.printAdded + !option.printDeleted + !option.printCommon) == 2)
 		option.needStartStop = false;
+
 #ifndef NO_MINUS_A
 	/* Not needed if there is no -a option, as option.transliterate will always
 	   be true then. */
-	if (TEST_BIT(option.whitespace, '\n')) {
+	if (!TEST_BIT(option.whitespace, '\n'))
 		option.transliterate = true;
-	}
 #endif
 END_FUNCTION
